@@ -56,16 +56,12 @@
 #   error Uho! This should not happen!
 #endif
 
-struct gprs_context_data {
+struct context_data {
 	GIsiClient *client;
 	GIsiModem *idx;
 	uint16_t gpds;	/* GPDS object handle */
-	GSList *contexts;
-};
-
-struct context_data {
 	unsigned cid;	/* oFono core context ID */
-	struct ofono_gprs_context *driver;
+	struct ofono_gprs_context *context;
 	union {
 		ofono_gprs_context_up_cb_t up_cb;
 		ofono_gprs_context_cb_t down_cb;
@@ -74,6 +70,7 @@ struct context_data {
 
 	GIsiPEP *pep;
 	GIsiPipe *pipe;
+	guint reset;
 
 	char apn[GPDS_MAX_APN_STRING_LENGTH + 1];
 	char username[GPDS_MAX_USERNAME_LENGTH + 1];
@@ -83,122 +80,131 @@ struct context_data {
 	uint8_t type;
 };
 
-static struct context_data *find_context_by_cid(GSList *contexts,
-						unsigned int cid)
+static gboolean client_reset(gpointer data)
 {
-	GSList *m = NULL;
+	struct context_data *cd = data;
 
-	for (m = contexts; m; m = m->next) {
-		struct context_data *cd = m->data;
+	g_isi_client_reset(cd->client);
+	cd->reset = 0;
 
-		if (cd->cid == cid)
-			return cd;
-	}
-	return NULL;
+	return FALSE;
 }
 
-static struct context_data *find_context_by_handle(GSList *contexts,
-							uint8_t handle)
+static void reset_context(struct context_data *cd)
 {
-	GSList *m = NULL;
-
-	for (m = contexts; m; m = m->next) {
-		struct context_data *cd = m->data;
-
-		if (cd->handle == handle)
-			return cd;
-	}
-	return NULL;
-}
-
-static void destroy_context(struct context_data *cd)
-{
-	if (!cd)
+	if (cd == NULL)
 		return;
 
-	DBG("destroying %p (cid=%u)", cd, cd->cid);
-
 	if (cd->pipe)
 		g_isi_pipe_destroy(cd->pipe);
 
 	if (cd->pep)
 		g_isi_pep_destroy(cd->pep);
 
-	g_free(cd);
+	cd->pep = NULL;
+	cd->pipe = NULL;
+	cd->handle = INVALID_ID;
+
+	cd->reset = g_idle_add(client_reset, cd);
 }
 
-static gboolean gprs_up_fail(struct context_data *cd)
-{
-	struct ofono_gprs_context *gc = cd->driver;
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
+typedef void (*ContextFailFunc)(struct context_data *cd);
 
+static void gprs_up_fail(struct context_data *cd)
+{
 	CALLBACK_WITH_FAILURE(cd->up_cb, NULL, 0, NULL, NULL, NULL, NULL,
 				cd->data);
-
-	gcd->contexts = g_slist_remove(gcd->contexts, cd);
-	destroy_context(cd);
-	return TRUE;
+	reset_context(cd);
 }
 
-static gboolean gprs_down_fail(struct context_data *cd)
+static void gprs_down_fail(struct context_data *cd)
 {
-	struct ofono_gprs_context *gc = cd->driver;
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
-
 	CALLBACK_WITH_FAILURE(cd->down_cb, cd->data);
-
-	gcd->contexts = g_slist_remove(gcd->contexts, cd);
-	destroy_context(cd);
-	return TRUE;
+	reset_context(cd);
 }
 
-static gboolean check_resp(GIsiClient *client,
-				const uint8_t *restrict msg, size_t len,
-				uint_fast8_t cmd, struct context_data *cd)
+static gboolean check_resp(const GIsiMessage *msg, uint8_t id, size_t minlen,
+				struct context_data *cd,
+				ContextFailFunc fail_cb)
 {
-	if (!msg) {
-		DBG("ISI client error: %d", g_isi_client_error(client));
+	const uint8_t *data = g_isi_msg_data(msg);
+
+	if (g_isi_msg_error(msg) < 0) {
+		DBG("ISI message error: %d", g_isi_msg_error(msg));
+		goto error;
+	}
+
+	if (g_isi_msg_id(msg) != id)
+		return FALSE;
+
+	if (g_isi_msg_data_len(msg) < minlen) {
+		DBG("truncated message");
+		goto error;
+	}
+
+	if (cd->handle != INVALID_ID && data[0] != cd->handle)
+		return FALSE;
+
+	if (data[1] != GPDS_OK) {
+		DBG("context error: %s (0x%02"PRIx8")",
+			gpds_status_name(data[1]), data[1]);
+
+		if (minlen > 2)
+			DBG("  fail cause: %s (0x%02"PRIx8")",
+				gpds_isi_cause_name(data[2]), data[2]);
+
+		goto error;
+	}
+
+	return TRUE;
+
+error:
+	if (fail_cb)
+		fail_cb(cd);
+
+	return FALSE;
+}
+
+static gboolean check_ind(const GIsiMessage *msg, size_t minlen,
+				struct context_data *cd)
+
+{
+	const uint8_t *data = g_isi_msg_data(msg);
+
+	if (g_isi_msg_error(msg) < 0) {
+		DBG("ISI message error: %d", g_isi_msg_error(msg));
 		return FALSE;
 	}
 
-	if (len < 3) {
+	if (g_isi_msg_data_len(msg) < minlen) {
 		DBG("truncated message");
 		return FALSE;
 	}
 
-	if (msg[0] != cmd) {
-		DBG("unexpected message ID: %s (0x%02"PRIx8")",
-			gpds_message_id_name(msg[0]), msg[0]);
+	if (cd->handle != INVALID_ID && data[0] != cd->handle)
 		return FALSE;
-	}
 
-	if ((cd->handle != INVALID_ID && msg[1] != cd->handle)
-		|| (msg[1] == INVALID_ID)) {
-		DBG("invalid context ID: 0x%02"PRIx8, msg[1]);
-		return FALSE;
-	}
-
-	if (msg[2] != GPDS_OK) {
-		DBG("context creation error: %s (0x%02"PRIx8")",
-			gpds_status_name(msg[2]), msg[2]);
-
-		if (len > 3)
-			DBG("  fail cause: %s (0x%02"PRIx8")",
-				gpds_isi_cause_name(msg[3]), msg[3]);
-
-		return FALSE;
-	}
 	return TRUE;
 }
 
-static void activate_ind_cb(GIsiClient *client,
-				const void *restrict data, size_t len,
-				uint16_t object, void *opaque)
+static void deactivate_ind_cb(const GIsiMessage *msg, void *opaque)
 {
-	struct gprs_context_data *gcd = opaque;
-	struct context_data *cd;
+	struct context_data *cd = opaque;
+	const uint8_t *data = g_isi_msg_data(msg);
 
-	const unsigned char *msg = data;
+	if (!check_ind(msg, 2, cd))
+		return;
+
+	DBG("context deactivated: %s (0x%02"PRIx8")",
+		gpds_isi_cause_name(data[2]), data[2]);
+
+	ofono_gprs_context_deactivated(cd->context, cd->cid);
+	reset_context(cd);
+}
+
+static void activate_ind_cb(const GIsiMessage *msg, void *opaque)
+{
+	struct context_data *cd = opaque;
 	GIsiSubBlockIter iter;
 
 	char ifname[IF_NAMESIZE];
@@ -207,18 +213,12 @@ static void activate_ind_cb(GIsiClient *client,
 	char *sdns = NULL;
 	const char *dns[3];
 
-	if (!msg || len < 3 || msg[0] != GPDS_CONTEXT_ACTIVATE_IND)
+	if (!check_ind(msg, 2, cd))
 		return;
 
-	cd = find_context_by_handle(gcd->contexts, msg[1]);
-	if (!cd) {
-		DBG("unknown context: 0x%02"PRIx8, msg[1]);
-		return;
-	}
-
-	for (g_isi_sb_iter_init(&iter, msg, len, 3);
-		g_isi_sb_iter_is_valid(&iter);
-		g_isi_sb_iter_next(&iter)) {
+	for (g_isi_sb_iter_init(&iter, msg, 2);
+			g_isi_sb_iter_is_valid(&iter);
+			g_isi_sb_iter_next(&iter)) {
 
 		uint8_t *addr_value = NULL;
 		uint8_t addr_len = 0;
@@ -228,9 +228,12 @@ static void activate_ind_cb(GIsiClient *client,
 		/* TODO: IPv6 address support */
 
 		case GPDS_PDP_ADDRESS_INFO:
-			if (!g_isi_sb_iter_get_byte(&iter, &addr_len, 3)
-				|| !g_isi_sb_iter_get_data(&iter,
-						(void *)&addr_value, 4))
+
+			if (!g_isi_sb_iter_get_byte(&iter, &addr_len, 3))
+				goto error;
+
+			if (!g_isi_sb_iter_get_data(&iter, (void *)&addr_value,
+							4))
 				goto error;
 
 			ip = alloca(INET_ADDRSTRLEN);
@@ -239,9 +242,12 @@ static void activate_ind_cb(GIsiClient *client,
 			break;
 
 		case GPDS_PDNS_ADDRESS_INFO:
-			if (!g_isi_sb_iter_get_byte(&iter, &addr_len, 3)
-				|| !g_isi_sb_iter_get_data(&iter,
-						(void *)&addr_value, 4))
+
+			if (!g_isi_sb_iter_get_byte(&iter, &addr_len, 3))
+				break;
+
+			if (!g_isi_sb_iter_get_data(&iter, (void *)&addr_value,
+							4))
 				break;
 
 			pdns = alloca(INET_ADDRSTRLEN);
@@ -250,9 +256,12 @@ static void activate_ind_cb(GIsiClient *client,
 			break;
 
 		case GPDS_SDNS_ADDRESS_INFO:
-			if (!g_isi_sb_iter_get_byte(&iter, &addr_len, 3)
-				|| !g_isi_sb_iter_get_data(&iter,
-						(void *)&addr_value, 4))
+
+			if (!g_isi_sb_iter_get_byte(&iter, &addr_len, 3))
+				break;
+
+			if (!g_isi_sb_iter_get_data(&iter, (void *)&addr_value,
+							4))
 				break;
 
 			sdns = alloca(INET_ADDRSTRLEN);
@@ -283,38 +292,19 @@ error:
 	gprs_up_fail(cd);
 }
 
-static void activate_fail_ind_cb(GIsiClient *client,
-					const void *restrict data, size_t len,
-					uint16_t object, void *opaque)
+static void activate_fail_ind_cb(const GIsiMessage *msg, void *opaque)
 {
-	const unsigned char *msg = data;
-	struct gprs_context_data *gcd = opaque;
-	struct context_data *cd;
+	struct context_data *cd = opaque;
 
-	if (!msg || len < 3 || msg[0] != GPDS_CONTEXT_ACTIVATE_FAIL_IND)
+	if (!check_ind(msg, 2, cd))
 		return;
 
-	cd = find_context_by_handle(gcd->contexts, msg[1]);
-	if (cd == NULL) {
-		DBG("unknown context: 0x%02"PRIx8, msg[1]);
-		return;
-	}
-
 	gprs_up_fail(cd);
 }
 
-static gboolean context_activate_cb(GIsiClient *client,
-					const void *restrict data, size_t len,
-					uint16_t object, void *opaque)
+static void context_activate_cb(const GIsiMessage *msg, void *cd)
 {
-	struct context_data *cd = opaque;
-
-	if (!check_resp(client, data, len, GPDS_CONTEXT_ACTIVATE_RESP, cd))
-		return gprs_up_fail(cd);
-
-	/* TODO: Add timeout here in case indications never come */
-
-	return TRUE;
+	check_resp(msg, GPDS_CONTEXT_ACTIVATE_RESP, 6, cd, gprs_up_fail);
 }
 
 static void send_context_activate(GIsiClient *client, void *opaque)
@@ -327,24 +317,29 @@ static void send_context_activate(GIsiClient *client, void *opaque)
 		0,		/* sub blocks */
 	};
 
-	if (g_isi_request_make(client, msg, sizeof(msg), GPDS_TIMEOUT,
-				context_activate_cb, cd))
+	g_isi_client_ind_subscribe(client, GPDS_CONTEXT_ACTIVATE_IND,
+				activate_ind_cb, cd);
+	g_isi_client_ind_subscribe(client, GPDS_CONTEXT_ACTIVATE_FAIL_IND,
+				activate_fail_ind_cb, cd);
+	g_isi_client_ind_subscribe(client, GPDS_CONTEXT_DEACTIVATE_IND,
+				deactivate_ind_cb, cd);
+
+	if (g_isi_client_send(client, msg, sizeof(msg),
+				GPDS_CTX_ACTIVATE_TIMEOUT,
+				context_activate_cb, cd, NULL) != NULL)
 		g_isi_pipe_start(cd->pipe);
 	else
 		gprs_up_fail(cd);
 }
 
-static gboolean context_auth_cb(GIsiClient *client,
-				const void *restrict data, size_t len,
-				uint16_t object, void *opaque)
+static void context_auth_cb(const GIsiMessage *msg, void *opaque)
 {
 	struct context_data *cd = opaque;
 
-	if (!check_resp(client, data, len, GPDS_CONTEXT_AUTH_RESP, cd))
-		return gprs_up_fail(cd);
+	if (!check_resp(msg, GPDS_CONTEXT_AUTH_RESP, 2, cd, gprs_up_fail))
+		return;
 
-	send_context_activate(client, cd);
-	return TRUE;
+	send_context_activate(cd->client, cd);
 }
 
 static void send_context_authenticate(GIsiClient *client, void *opaque)
@@ -377,36 +372,30 @@ static void send_context_authenticate(GIsiClient *client, void *opaque)
 		{ cd->password, password_len },
 	};
 
-	if (!g_isi_request_vmake(client, iov, 4, GPDS_TIMEOUT,
-					context_auth_cb, cd))
+	if (g_isi_client_vsend(client, iov, 4, GPDS_TIMEOUT,
+				context_auth_cb, cd, NULL) == NULL)
 		gprs_up_fail(cd);
 }
 
-static gboolean context_conf_cb(GIsiClient *client,
-				const void *restrict data, size_t len,
-				uint16_t object, void *opaque)
+static void context_conf_cb(const GIsiMessage *msg, void *opaque)
 {
 	struct context_data *cd = opaque;
 
-	if (!check_resp(client, data, len, GPDS_CONTEXT_CONFIGURE_RESP, cd))
-		return gprs_up_fail(cd);
+	if (!check_resp(msg, GPDS_CONTEXT_CONFIGURE_RESP, 2, cd, gprs_up_fail))
+		return;
 
 	if (cd->username[0] != '\0')
-		send_context_authenticate(client, cd);
+		send_context_authenticate(cd->client, cd);
 	else
-		send_context_activate(client, cd);
-
-	return TRUE;
+		send_context_activate(cd->client, cd);
 }
 
-static gboolean link_conf_cb(GIsiClient *client,
-				const void *restrict data, size_t len,
-				uint16_t object, void *opaque)
+static void link_conf_cb(const GIsiMessage *msg, void *opaque)
 {
 	struct context_data *cd = opaque;
 	size_t apn_len = strlen(cd->apn);
 
-	const unsigned char msg[] = {
+	const unsigned char req[] = {
 		GPDS_CONTEXT_CONFIGURE_REQ,
 		cd->handle,	/* context ID */
 		cd->type,	/* PDP type */
@@ -423,103 +412,75 @@ static gboolean link_conf_cb(GIsiClient *client,
 	};
 
 	const struct iovec iov[2] = {
-		{ (uint8_t *)msg, sizeof(msg) },
+		{ (uint8_t *)req, sizeof(req) },
 		{ cd->apn, apn_len },
 	};
 
-	if (!check_resp(client, data, len, GPDS_LL_CONFIGURE_RESP, cd))
-		return gprs_up_fail(cd);
+	if (!check_resp(msg, GPDS_LL_CONFIGURE_RESP, 2, cd, gprs_up_fail))
+		return;
 
-	if (!g_isi_request_vmake(client, iov, 2, GPDS_TIMEOUT,
-					context_conf_cb, cd))
-		return gprs_up_fail(cd);
-
-	return TRUE;
+	if (g_isi_client_vsend(cd->client, iov, 2, GPDS_TIMEOUT,
+				context_conf_cb, cd, NULL) == NULL)
+		gprs_up_fail(cd);
 }
 
-static gboolean create_context_cb(GIsiClient *client,
-					const void *restrict data, size_t len,
-					uint16_t object, void *opaque)
+static void create_context_cb(const GIsiMessage *msg, void *opaque)
 {
-	const unsigned char *resp = data;
 	struct context_data *cd = opaque;
+	const uint8_t *data = g_isi_msg_data(msg);
 
-	unsigned char msg[] = {
+	unsigned char req[] = {
 		GPDS_LL_CONFIGURE_REQ,
 		0x00,		/* GPDS context ID, added later */
 		g_isi_pipe_get_handle(cd->pipe),
 		GPDS_LL_PLAIN,	/* link type */
 	};
 
-	if (!check_resp(client, data, len, GPDS_CONTEXT_ID_CREATE_RESP, cd))
-		return gprs_up_fail(cd);
+	if (!check_resp(msg, GPDS_CONTEXT_ID_CREATE_RESP, 2, cd, gprs_up_fail))
+		return;
 
-	cd->handle = msg[1] = resp[1];
+	cd->handle = req[1] = data[0];
 
-	if (!g_isi_request_make(client, msg, sizeof(msg), GPDS_TIMEOUT,
-				link_conf_cb, cd))
-		return gprs_up_fail(cd);
-
-	/* TODO: send context configuration at the same time? */
-
-	return TRUE;
+	if (g_isi_client_send(cd->client, req, sizeof(req), GPDS_TIMEOUT,
+				link_conf_cb, cd, NULL) == NULL)
+		gprs_up_fail(cd);
 }
 
 static void create_pipe_cb(GIsiPipe *pipe)
 {
 	struct context_data *cd = g_isi_pipe_get_userdata(pipe);
-	struct ofono_gprs_context *gc = cd->driver;
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 
 	const unsigned char msg[] = {
 		GPDS_CONTEXT_ID_CREATE_REQ,
 	};
 
-	if (!g_isi_request_make(gcd->client, msg, sizeof(msg), GPDS_TIMEOUT,
-				create_context_cb, cd))
+	if (g_isi_client_send(cd->client, msg, sizeof(msg), GPDS_TIMEOUT,
+				create_context_cb, cd, NULL) == NULL)
 		gprs_up_fail(cd);
 }
 
-static void deactivate_ind_cb(GIsiClient *client,
-				const void *restrict data, size_t len,
-				uint16_t object, void *opaque)
-{
-	struct gprs_context_data *gcd = opaque;
-	struct context_data *cd;
-
-	const unsigned char *msg = data;
-
-	if (!msg || len < 3 || msg[0] != GPDS_CONTEXT_DEACTIVATE_IND)
-		return;
-
-	cd = find_context_by_handle(gcd->contexts, msg[1]);
-	if (cd == NULL) {
-		DBG("unknown context: 0x%02"PRIx8, msg[1]);
-		return;
-	}
-
-	DBG("context deactivated: %s (0x%02"PRIx8")",
-		gpds_isi_cause_name(msg[3]), msg[3]);
-
-	ofono_gprs_context_deactivated(cd->driver, cd->cid);
-
-	gcd->contexts = g_slist_remove(gcd->contexts, cd);
-	destroy_context(cd);
-}
-
 static void isi_gprs_activate_primary(struct ofono_gprs_context *gc,
 				const struct ofono_gprs_primary_context *ctx,
 				ofono_gprs_context_up_cb_t cb, void *data)
 {
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
-	struct context_data *cd = g_try_new0(struct context_data, 1);
-	struct context_data *old = NULL;
+	struct context_data *cd = ofono_gprs_context_get_data(gc);
 
-	if (!cd)
+	DBG("activate: gpds = 0x%04x", cd->gpds);
+
+	if (cd == NULL || !cd->gpds) {
+		/* GPDS is not reachable */
+		CALLBACK_WITH_FAILURE(cb, NULL, 0, NULL, NULL, NULL,
+					NULL, data);
 		return;
+	}
+
+	if (cd->reset) {
+		g_isi_client_reset(cd->client);
+		g_source_remove(cd->reset);
+		cd->reset = 0;
+	}
 
 	cd->cid = ctx->cid;
-	cd->driver = gc;
 	cd->up_cb = cb;
 	cd->data = data;
 	cd->pep = NULL;
@@ -527,17 +488,9 @@ static void isi_gprs_activate_primary(struct ofono_gprs_context *gc,
 	cd->handle = INVALID_ID;
 	cd->type = GPDS_PDP_TYPE_IPV4;
 
-	old = find_context_by_cid(gcd->contexts, ctx->cid);
-	if (old) {
-		DBG("duplicate context: %u", ctx->cid);
-		goto error;
-	}
-
-	gcd->contexts = g_slist_append(gcd->contexts, cd);
-
 	if (strlen(ctx->apn) >= GPDS_MAX_APN_STRING_LENGTH
-		|| strlen(ctx->username) >= GPDS_MAX_USERNAME_LENGTH
-		|| strlen(ctx->password) >= GPDS_MAX_PASSWORD_LENGTH)
+			|| strlen(ctx->username) >= GPDS_MAX_USERNAME_LENGTH
+			|| strlen(ctx->password) >= GPDS_MAX_PASSWORD_LENGTH)
 		goto error;
 
 	strncpy(cd->apn, ctx->apn, GPDS_MAX_APN_STRING_LENGTH);
@@ -549,13 +502,13 @@ static void isi_gprs_activate_primary(struct ofono_gprs_context *gc,
 	strncpy(cd->password, ctx->password, GPDS_MAX_PASSWORD_LENGTH);
 	cd->username[GPDS_MAX_PASSWORD_LENGTH] = '\0';
 
-	cd->pep = g_isi_pep_create(gcd->idx, NULL, NULL);
+	cd->pep = g_isi_pep_create(cd->idx, NULL, NULL);
 	if (cd->pep == NULL)
 		goto error;
 
-	cd->pipe = g_isi_pipe_create(gcd->idx, create_pipe_cb,
+	cd->pipe = g_isi_pipe_create(cd->idx, create_pipe_cb,
 					g_isi_pep_get_object(cd->pep),
-					gcd->gpds, PN_PEP_TYPE_GPRS,
+					cd->gpds, PN_PEP_TYPE_GPRS,
 					PN_PEP_TYPE_GPRS);
 	if (cd->pipe == NULL)
 		goto error;
@@ -567,127 +520,101 @@ error:
 	gprs_up_fail(cd);
 }
 
-static gboolean context_deactivate_cb(GIsiClient *client,
-					const void *restrict data, size_t len,
-					uint16_t object,
-					void *opaque)
+static void context_deactivate_cb(const GIsiMessage *msg, void *opaque)
 {
 	struct context_data *cd = opaque;
-	struct ofono_gprs_context *gc = cd->driver;
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 
-	if (!check_resp(client, data, len, GPDS_CONTEXT_DEACTIVATE_RESP, cd))
-		return gprs_down_fail(cd);
-
-	gcd->contexts = g_slist_remove(gcd->contexts, cd);
+	if (!check_resp(msg, GPDS_CONTEXT_DEACTIVATE_RESP, 2, cd,
+			gprs_down_fail))
+		return;
 
 	CALLBACK_WITH_SUCCESS(cd->down_cb, cd->data);
-	destroy_context(cd);
-
-	return TRUE;
+	reset_context(cd);
 }
 
 static void isi_gprs_deactivate_primary(struct ofono_gprs_context *gc,
 					unsigned int cid,
 					ofono_gprs_context_cb_t cb, void *data)
 {
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
-	struct context_data *cd;
+	struct context_data *cd = ofono_gprs_context_get_data(gc);
 
 	unsigned char msg[] = {
 		GPDS_CONTEXT_DEACTIVATE_REQ,
 		0x00,	/* GPDS context ID, added later */
 	};
 
-	cd = find_context_by_cid(gcd->contexts, cid);
-	if (!cd) {
-		DBG("unknown context: %u", cid);
+	if (cd == NULL)
 		return;
-	}
 
 	cd->down_cb = cb;
 	cd->data = data;
 
 	msg[1] = cd->handle;
 
-	if (!g_isi_request_make(gcd->client, msg, sizeof(msg), GPDS_TIMEOUT,
-				context_deactivate_cb, cd))
+	if (g_isi_client_send(cd->client, msg, sizeof(msg),
+				GPDS_CTX_DEACTIVATE_TIMEOUT,
+				context_deactivate_cb, cd, NULL) == NULL) {
 		gprs_down_fail(cd);
+		return;
+	}
 }
 
-static void gpds_ctx_reachable_cb(GIsiClient *client, gboolean alive,
-					uint16_t object,
-					void *opaque)
+static void gpds_ctx_reachable_cb(const GIsiMessage *msg, void *opaque)
 {
-	struct ofono_gprs_context *gc = opaque;
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
-	const char *debug;
+	struct context_data *cd = opaque;
 
-	if (!alive) {
+	if (g_isi_msg_error(msg) < 0) {
 		DBG("unable to bootstrap gprs context driver");
 		return;
 	}
 
-	DBG("%s (v%03d.%03d) for PDP contexts",
-		pn_resource_name(g_isi_client_resource(client)),
-		g_isi_version_major(client),
-		g_isi_version_minor(client));
-
-	gcd->gpds = object;
-
-	debug = getenv("OFONO_ISI_DEBUG");
-	if (debug && (strcmp(debug, "all") == 0 || strcmp(debug, "gpds") == 0))
-		g_isi_client_set_debug(gcd->client, gpds_debug, NULL);
-
-	g_isi_subscribe(client, GPDS_CONTEXT_ACTIVATE_IND,
-			activate_ind_cb, gcd);
-	g_isi_subscribe(client, GPDS_CONTEXT_ACTIVATE_FAIL_IND,
-			activate_fail_ind_cb, gcd);
-	g_isi_subscribe(client, GPDS_CONTEXT_DEACTIVATE_IND,
-			deactivate_ind_cb, gcd);
+	cd->gpds = g_isi_msg_object(msg);
 }
 
 static int isi_gprs_context_probe(struct ofono_gprs_context *gc,
 					unsigned int vendor, void *user)
 {
 	GIsiModem *idx = user;
-	struct gprs_context_data *gcd = g_try_new0(struct gprs_context_data, 1);
+	struct context_data *cd = g_try_new0(struct context_data, 1);
 
-	if (!gcd)
+	if (cd == NULL)
 		return -ENOMEM;
 
-	gcd->client = g_isi_client_create(idx, PN_GPDS);
-	if (!gcd->client) {
-		g_free(gcd);
+	cd->client = g_isi_client_create(idx, PN_GPDS);
+	if (cd->client == NULL) {
+		g_free(cd);
 		return -ENOMEM;
 	}
 
-	ofono_gprs_context_set_data(gc, gcd);
+	cd->idx = idx;
+	cd->context = gc;
+	ofono_gprs_context_set_data(gc, cd);
 
-	gcd->idx = idx;
-	gcd->contexts = NULL;
-
-	g_isi_verify(gcd->client, gpds_ctx_reachable_cb, gc);
+	g_isi_client_verify(cd->client, gpds_ctx_reachable_cb, cd, NULL);
 
 	return 0;
 }
 
 static void isi_gprs_context_remove(struct ofono_gprs_context *gc)
 {
-	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
-	GSList *m;
+	struct context_data *cd = ofono_gprs_context_get_data(gc);
 
 	ofono_gprs_context_set_data(gc, NULL);
 
-	for (m = gcd->contexts; m; m = m->next)
-		destroy_context(m->data);
+	if (cd == NULL)
+		return;
 
-	g_slist_free(gcd->contexts);
+	if (cd->reset)
+		g_source_remove(cd->reset);
 
-	if (gcd->client)
-		g_isi_client_destroy(gcd->client);
+	if (cd->pipe != NULL)
+		g_isi_pipe_destroy(cd->pipe);
 
-	g_free(gcd);
+	if (cd->pep != NULL)
+		g_isi_pep_destroy(cd->pep);
+
+	g_isi_client_destroy(cd->client);
+	g_free(cd);
 }
 
 static struct ofono_gprs_context_driver driver = {
